from importlib import metadata

CATWALK_CLIENT_VERSION = metadata.version("catwalk_client")

CATWALK_AUTH_HEADER = "Catwalk-Authorization"
CATWALK_CLIENT_LOCATION = "Client-Location"
CATWALK_USER_AGENT_HEADER_VALUE = f"Catwalk-Client/{CATWALK_CLIENT_VERSION}"
CATWALK_CLIENT_THREAD_NAME_PREFIX = "catwalk_client_"

CATWALK_CASE_TRACK_ID_ENDPOINT = "/api/cases/track-id/%s"
CATWALK_GET_CASE_EVALUATIONS_BY_TRACK_ID_ENDPOINT = "/api/evaluation/case/track-id/%s"
CATWALK_CREATE_SESSION_ENDPOINT = "/api/evaluation/sessions"
CATWALK_ADD_CASES_TO_SESSION_ENDPOINT = "/api/evaluation/sessions/%s/cases"
CATWALK_START_SESSION_ENDPOINT = "/api/evaluation/sessions/%s/start"
CATWALK_GET_USERS_ENDPOINT = "/api/auth/users"
CATWALK_CLIENT_LOG_HANDLER = "catwalk_client_log_handler"
CATWALK_CLIENT_LOG_NAME = "catwalk_client"



from typing import List, Dict

from pydantic import BaseModel


class AttributeNames(BaseModel):
    """Attribute names used inside case details."""

    context: List[str] = []
    query: List[str] = []
    response: List[str] = []


class SessionConfig(BaseModel):
    """Evaluation session configuration.

    `show_responses` - if `False` the `response` case detail
    won't be shown during the evaluation.

    `allow_user_evaluations` - if `True` the evaluator will
    be able to add evaluations to the cases during the evaluation.

    `submitters_disabled_fields` - case details with included
    attribute names won't be displayed to the evaluator during the
    evaluation. Session submitters configuration type. Attribute
    names used by submitters' cases. First level dictionary keys
    are submitter names and second level dictionary keys are
    submitter versions.
    """

    show_responses: bool = True
    allow_user_evaluations: bool = True
    submitters_disabled_fields: Dict[str, Dict[str, AttributeNames]] = {}



from __future__ import annotations

import time
from json import loads
from datetime import datetime
from typing import Literal

from catwalk_common import OpenCase

from catwalk_client.common.catwalk_http_client import CatwalkHTTPClient
from catwalk_client.common.exception import CatwalkClientException
from catwalk_client.common.logger import init_logger

log = init_logger()

SORT_ORDER = Literal["asc", "desc"]


class CaseExporter:
    def __init__(
        self,
        http_client: CatwalkHTTPClient,
        submitter_name: str | None = None,
        submitter_version: str | None = None,
        datetime_from: datetime | None = None,
        datetime_to: datetime = datetime.now(),
        sort_by: list[str] | str | None = None,
        sort_order: list[SORT_ORDER] | SORT_ORDER = "desc",
        limit: int = 100,
        offset: int = 0,
        max_retries: int = 5,
    ):
        """Create CaseExporter object (EXPORT V2).

        Args:
            submitter_name (Optional[str], optional): Name of submitter to export. Defaults to None.
            submitter_version (Optional[str], optional): Version of submitter to export. Defaults to None.
            datetime_from (Optional[datetime], optional): Lower boundary of case creation date. Defaults to None.
            datetime_to (datetime, optional): Upper boundary of case creation date. Defaults to current time.
            sort_by (Union[list[str], str, None], optional): Field(s) by which exported cases will be ordered. Defaults to None.
            sort_order (Union[list[SORT_ORDER], SORT_ORDER], optional): Order(s) by which exported cases will be ordered. Defaults to "desc".
            limit (int, optional): Limit of a single batch fetched from the API (not overall limit).
                                   Defaults to 100. Max is 1000, if the limit is higher than 1000 the API will return an error.
            offset (int, optional): Starting offset of fetched data. Defaults to 0.
            max_retries (int, optional): Maximum number of retries in case the request does not succeed. Defaults to 5.
        """
        self.path = "/api/v2/cases/export"
        self.http_client = http_client
        self.max_retries = max_retries

        self.submitter_name = submitter_name
        self.submitter_version = submitter_version
        self.datetime_from = datetime_from
        self.datetime_to = datetime_to
        self.sort_by = sort_by
        self.sort_order = sort_order
        self.limit = limit
        self.offset = offset

    def export(self):
        """Fetch cases in form of a generator where each element of is a case.

        Yields:
            OpenCase: Cases returned by the API.
        """
        current_retry = 0
        while True:
            response, success, status_code = self.http_client.get(
                self.path, self._get_query_params()
            )

            retry = status_code < 0 or status_code >= 500
            if not success and retry:
                self._handle_retry(current_retry, response)
                current_retry += 1
                continue
            elif not success:
                raise CatwalkClientException(response)

            current_retry = 0
            data = loads(response)

            for case in data["cases"]:
                yield OpenCase.parse_obj(case)

            if len(data["cases"]) < self.limit:
                break

            self.offset += self.limit

    def _handle_retry(self, retry: int, response: str):
        """Check if the unsuccessful call can be retried and apply delay to the next call.

        Args:
            retry (int): Current retry.
            response (str): Response of an unsuccessful call.

        Raises:
            CatwalkClientException: When the maximum number of retries is exceeded.
        """
        if retry < self.max_retries:
            multiplier = (retry * 2) or 1
            log.warning(
                f"Retry ({retry + 1}/{self.max_retries}) because got error {response}"
            )
            time.sleep(0.5 * multiplier)
        else:
            raise CatwalkClientException("Max retries exceeded")

    def _get_query_params(self):
        """Return query params in form of a list of tuples.

        Returns:
            list[tuple[str, Any]]: List of tuples where first element is a query param key and the second is its value.
        """
        query_params = [
            (k, v)
            for k, v in {
                "submitter_name": self.submitter_name,
                "submitter_version": self.submitter_version,
                "datetime_from": self.datetime_from.isoformat(),
                "datetime_to": self.datetime_to.isoformat(),
                "limit": self.limit,
                "offset": self.offset,
            }.items()
            if v is not None
        ]

        if self.sort_by:
            query_params += [
                ("sort_by", sb) for sb in self._parse_list_query_param(self.sort_by)
            ]

        if self.sort_order:
            query_params += [
                ("sort_order", so)
                for so in self._parse_list_query_param(self.sort_order)
            ]

        return query_params

    def _parse_list_query_param(self, list_param: list | str) -> list[str]:
        """Check if given value is a string, if yes return it as a list with a single string.

        Args:
            list_param (Union[list, str]): Value to parse.

        Returns:
            list[str]: Parsed given value to list of strings.
        """
        param = list_param
        if not isinstance(list_param, list):
            param = [list_param]
        return param



import logging

from catwalk_client.common.constants import (
    CATWALK_CLIENT_LOG_HANDLER,
    CATWALK_CLIENT_LOG_NAME,
)


class CatwalkClientLoggingFormatter(logging.Formatter):
    GRAY = "\x1b[38;20m"
    BLUE = "\x1b[34;20m"
    YELLOW = "\x1b[33;20m"
    RED = "\x1b[31;20m"
    BOLD_RED = "\x1b[31;1m"
    RESET = "\x1b[0m"

    fmt = "[%(asctime)s] [%(levelname)s] [%(name)s]: %(message)s"
    fmt_more = "[%(asctime)s] [%(levelname)s] [%(name)s] (%(filename)s:%(lineno)d): %(message)s"
    date_fmt = "%Y-%m-%d %H:%M:%S %z"

    FORMATS = {
        logging.DEBUG: BLUE + fmt_more + RESET,
        logging.INFO: GRAY + fmt + RESET,
        logging.WARNING: YELLOW + fmt + RESET,
        logging.ERROR: RED + fmt_more + RESET,
        logging.CRITICAL: BOLD_RED + fmt_more + RESET,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt=self.date_fmt)
        return formatter.format(record)


def init_logger(logger_name: str = CATWALK_CLIENT_LOG_NAME):
    log = logging.getLogger(logger_name)

    if any(
        [handler.get_name() == CATWALK_CLIENT_LOG_HANDLER for handler in log.handlers]
    ):
        log.debug(f"Logger '{logger_name}' has already been initialized!")
        return log

    log_handler = logging.StreamHandler()
    log_handler.set_name(CATWALK_CLIENT_LOG_HANDLER)
    log_handler.setLevel(logging.DEBUG)
    log_handler.setFormatter(CatwalkClientLoggingFormatter())
    log.addHandler(log_handler)

    log_level = "INFO"
    log.setLevel(log_level)
    log.debug(f"Log level: {log_level}")

    return log


import logging



from __future__ import annotations

from typing import Any


class CaseBuilder:
    def __init__(self, client):
        self._client = client
        self._track_id = None
        self._query: list[dict] = []
        self._context: list[dict] = []
        self._response: list[dict] = []
        self._metadata: dict = {}

    def set_track_id(self, track_id: str):
        self._track_id = track_id
        return self

    def add_query(self, name: str, value: Any, type: str | dict):
        self._query.append({"name": name, "value": value, "type": type})
        return self

    def add_context(self, name: str, value: Any, type: str | dict):
        self._context.append({"name": name, "value": value, "type": type})
        return self

    def add_response(
        self,
        name: str,
        value: Any,
        type: str | dict,
    ):
        self._response.append({"name": name, "value": value, "type": type})
        return self

    def set_metadata(self, **kwargs):
        self._metadata = kwargs
        return self

    def send(self):
        self._client.send(
            {
                "track_id": self._track_id,
                "query": self._query,
                "context": self._context,
                "response": self._response,
                "metadata": self._metadata,
            }
        )

    def update(self):
        """Update the case details (`query`, `context`, `response`, `metadata`)."""
        data = {
            "query": self._query,
            "context": self._context,
            "response": self._response,
            "metadata": self._metadata,
        }

        self._client.update(self._track_id, data)



from catwalk_client.common.logger import init_logger

log = init_logger()


class CatwalkClientException(Exception):
    def __init__(self, reason: str, **extra):
        log.error(reason)
        log.debug(extra)
        log.debug(self.__traceback__)

        self.reason = reason
        [self.__setattr__(name, value) for name, value in extra.items()]



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from json import loads

from catwalk_common import CommonCaseFormat, OpenCase, EvaluationResult
from pydantic import error_wrappers

from catwalk_client.common import CatwalkHTTPClient
from catwalk_client.common.constants import (
    CATWALK_ADD_CASES_TO_SESSION_ENDPOINT,
    CATWALK_CLIENT_THREAD_NAME_PREFIX,
    CATWALK_CASE_TRACK_ID_ENDPOINT,
    CATWALK_CREATE_SESSION_ENDPOINT,
    CATWALK_GET_CASE_EVALUATIONS_BY_TRACK_ID_ENDPOINT,
    CATWALK_CLIENT_LOG_NAME,
    CATWALK_GET_USERS_ENDPOINT,
    CATWALK_START_SESSION_ENDPOINT,
)
from catwalk_client.common.logger import init_logger
from catwalk_client.common._session import SessionConfig
from catwalk_client.tools import CaseBuilder, CaseExporter, CaseExporterV2, SORT_ORDER


class CatwalkClient:
    log = init_logger(CATWALK_CLIENT_LOG_NAME)

    def __init__(
        self,
        submitter_name: str | None = None,
        submitter_version: str | None = None,
        catwalk_url: str | None = None,
        auth_token: str | None = None,
        insecure: bool = True,
        timeout: int = 30,
        timezone: str = "UTC",
        concurrent: bool = False,
        max_workers: int = 4,
    ):
        self.http_client = CatwalkHTTPClient(
            catwalk_url, auth_token, insecure, timeout, timezone
        )
        self.submitter_name = submitter_name
        self.submitter_version = submitter_version
        self._max_workers = max_workers
        self._concurrent = concurrent

        if self._concurrent:
            self._executor = ThreadPoolExecutor(
                max_workers=self._max_workers,
                thread_name_prefix=CATWALK_CLIENT_THREAD_NAME_PREFIX,
            )
        else:
            self._executor = None

    def __setattr__(self, __name, __value):
        if __name != "http_client" and __name in self.http_client.__dict__.keys():
            self.log.warn(
                f" [DEPRECATED] Usage of 'CatwalkClient.{__name}=<value>' is DEPRECATED. Please use 'set_{__name}' method!"
            )
            self.http_client.__dict__[__name] = __value

        self.__dict__[__name] = __value

    def set_auth_token(self, auth_token: str):
        self.http_client.auth_token = auth_token

    def set_catwalk_url(self, catwalk_url: str):
        self.http_client.url = catwalk_url

    def set_insecure(self, insecure: bool):
        self.http_client.insecure = insecure

    def new_case(self) -> CaseBuilder:
        return CaseBuilder(client=self)

    def alter_case(self, track_id: str) -> CaseBuilder | None:
        """Get CaseBuilder object filled with details of desired case. It fetches
        the case with a given `track_id` and fills CaseBuilder fields with fetched
        case's query, context, response and metadata information.

        Args:
            track_id (str): Track ID of a case.

        Returns:
            CaseBuilder | None: Object filled with case details of a case with given track ID.
            Returns None on an unsuccessful case fetch operation.
        """
        case = self.get_case(track_id)

        if not case:
            return None

        case_builder = self.new_case()
        case_builder._track_id = case.track_id
        case_builder._query = [query.dict() for query in case.query]
        case_builder._context = [context.dict() for context in case.context]
        case_builder._response = [response.dict() for response in case.response]
        case_builder._metadata = case.metadata

        return case_builder

    def send(self, case: dict):
        if self._concurrent and self._executor:
            self._executor.submit(self._send, case)
        else:
            self._send(case)

    def _send(self, case: dict):
        try:
            case: CommonCaseFormat = CommonCaseFormat(
                submitter={
                    "name": self.submitter_name,
                    "version": self.submitter_version,
                },
                **case,
            )

            response, success, _ = self.http_client.post(
                "/api/cases/collect", case.dict()
            )

            if success:
                case_id = loads(response)["id"]
                self.log.info(f"Collected catwalk case: {case_id}")
            else:
                self.log.error(f"Couldn't collect the case.\n{response}")
        except error_wrappers.ValidationError as ex:
            self.log.error(f"ValidationError: \n{ex}")
        except Exception as ex:
            self.log.error(f"{type(ex).__name__}: \n{str(ex)}\n{ex.__traceback__}")

    def export_cases(
        self,
        from_datetime: datetime,
        to_datetime: datetime,
        submitter_name: str | None = None,
        submitter_version: str | None = None,
        max_retries: int = 5,
        limit: int | None = None,
    ):
        exporter = CaseExporter(
            http_client=self.http_client,
            from_datetime=from_datetime,
            to_datetime=to_datetime,
            submitter_name=submitter_name or self.submitter_name,
            submitter_version=submitter_version or self.submitter_version,
            max_retries=max_retries,
            limit=limit,
        )
        yield from exporter.export()

    def export_cases_v2(
        self,
        submitter_name: str | None = None,
        submitter_version: str | None = None,
        datetime_from: datetime | None = None,
        datetime_to: datetime = datetime.now(),
        sort_by: list[str] | str | None = None,
        sort_order: list[SORT_ORDER] | SORT_ORDER = "desc",
        limit: int = 100,
        offset: int = 0,
        max_retries: int = 5,
    ):
        """Export cases using the v2 export endpoint.

        Args:
            submitter_name (Optional[str], optional): Name of submitter to export. Defaults to None.
            submitter_version (Optional[str], optional): Version of submitter to export. Defaults to None.
            datetime_from (Optional[datetime], optional): Lower boundary of case creation date. Defaults to None.
            datetime_to (datetime, optional): Upper boundary of case creation date. Defaults to current time.
            sort_by (Union[list[str], str, None], optional): Field(s) by which exported cases will be ordered. Defaults to None.
            sort_order (Union[list[SORT_ORDER], SORT_ORDER], optional): Order(s) by which exported cases will be ordered. Defaults to "desc".
            limit (int, optional): Size limit of a single batch fetched from the API (not overall limit).
                                   Defaults to 100. Max is 1000, if the limit is higher than 1000 the API will return an error.
            offset (int, optional): Starting offset of fetched data. Defaults to 0.
            max_retries (int, optional): Maximum number of retries in case the request does not succeed. Defaults to 5.

        Yields:
            OpenCase: Cases returned by the API.
        """
        exporter = CaseExporterV2(
            http_client=self.http_client,
            submitter_name=submitter_name or self.submitter_name,
            submitter_version=submitter_version or self.submitter_version,
            datetime_from=datetime_from,
            datetime_to=datetime_to,
            sort_by=sort_by,
            sort_order=sort_order,
            limit=limit,
            offset=offset,
            max_retries=max_retries,
        )
        yield from exporter.export()

    def get_case(self, track_id: str) -> OpenCase | None:
        """Get case that's track_id matches a given string.

        Args:
            track_id (str): Case's track_id.

        Returns:
            OpenCase: Case in a Common Case Format with some additional information
            like creation date, id, status and archivization status. Returns None
            on HTTP error.
        """
        path = CATWALK_CASE_TRACK_ID_ENDPOINT % track_id
        response, success, _ = self.http_client.get(path)

        if success:
            case = OpenCase.parse_obj(loads(response))
            return case
        else:
            self.log.info(f"Couldn't fetch the case.\n{response}")
            return None

    def get_case_evaluation_results(
        self, track_id: str
    ) -> list[EvaluationResult] | None:
        """Get evaluation results of a case that's track_id matches a given string.

        Args:
            track_id (str): Case's track_id.

        Returns:
            list[EvaluationResult] | None: List of evaluations of a given case. Returns
            None on HTTP error.
        """
        path = CATWALK_GET_CASE_EVALUATIONS_BY_TRACK_ID_ENDPOINT % track_id
        response, success, _ = self.http_client.get(path)

        if success:
            evaluations = loads(response)["evaluations"]
            return [EvaluationResult.parse_obj(ev) for ev in evaluations]
        else:
            self.log.info(f"Couldn't fetch evaluation results.\n{response}")
            return None

    def update(self, track_id: str, case: dict):
        """Send given case details in order to update a case. `track_id` is
        required in order to update a case.

        Args:
            track_id (str): Track ID of a case.
            case (dict): New case details (`query`, `context`, `response`, `metadata`).
        """
        try:
            case: CommonCaseFormat = CommonCaseFormat(
                submitter={
                    "name": self.submitter_name,
                    "version": self.submitter_version,
                },
                **case,
            )

            path = CATWALK_CASE_TRACK_ID_ENDPOINT % track_id
            response, success, _ = self.http_client.patch(path, case.dict())

            if success:
                track_id = loads(response)["track_id"]
                self.log.info(f" Updated case with track_id: {track_id}")
            else:
                self.log.error(f" Couldn't update the case. {response}")
        except error_wrappers.ValidationError as ex:
            self.log.error(f" ValidationError: \n{ex}")
        except Exception as ex:
            self.log.error(f" {type(ex).__name__}: \n{str(ex)}\n{ex.__traceback__}")

    def create_session(self, name: str, description: str = "") -> str | None:
        """Create a Catwalk evaluation session.

        Args:
            name (str): Name of a session.
            description (str, optional): Session description. Defaults to "".

        Returns:
            str | None: Created session ID. Returns `None` when there was an
            error while creating a session.
        """
        payload = {
            "name": name,
            "description": description,
        }
        response, success, _ = self.http_client.post(
            CATWALK_CREATE_SESSION_ENDPOINT, payload
        )

        if not success:
            self.log.error(
                f"There was an error while trying to create a session.\n{response}"
            )
            return None

        session_id = loads(response)["session_id"]
        self.log.info(f'Created session "{name}" with ID: {session_id}')

        return session_id

    def add_cases_to_session(
        self, session_id: str, track_ids: list[str]
    ) -> list[str] | None:
        """Add cases to the session using track IDs.

        Args:
            session_id (str): ID of a session.
            track_ids (list[str]): List of cases' track IDs.

        Returns:
            list[str] | None: List of IDs of created session cases.
            Returns `None` when there was an error while assigning
            cases to the session.
        """
        payload = {"track_ids": track_ids}
        path = CATWALK_ADD_CASES_TO_SESSION_ENDPOINT % session_id
        response, success, _ = self.http_client.post(path, payload)

        if not success:
            self.log.error(
                f"There was an error while trying to add cases to the session.\n{response}"
            )
            return None

        session_case_ids = loads(response)["session_case_ids"]
        self.log.info(
            f"Added {len(session_case_ids)} of {len(track_ids)} cases to session with ID: {session_id}"
        )

        return session_case_ids

    def start_session(
        self,
        session_id: str,
        assign_to: list[str] = [],
        session_config: SessionConfig = SessionConfig(),
    ) -> bool:
        """Start a session.

        Args:
            session_id (str): ID of a session.
            assign_to (list[str]): List of users' emails that will be
            assigned as the session evaluators. Session is always assigned
            to the owner, even when his email is not provided. Defaults to
            [].
            session_config (SessionConfig, optional): Session configuration.
            Defaults to SessionConfig() (default configuration).

        Returns:
            bool: Boolean value that informs whether the session has
            been successfully started.
        """
        response, success, _ = self.http_client.get(CATWALK_GET_USERS_ENDPOINT)

        if not success:
            self.log.error(f"There was an error while fetching users list.\n{response}")
            return False

        users = loads(response)["users"]
        users = [user["id"] for user in users if user["email"] in assign_to]

        payload = {"assigned_users": users, "config": session_config.dict()}
        path = CATWALK_START_SESSION_ENDPOINT % session_id
        response, success, _ = self.http_client.post(path, payload)

        if not success:
            self.log.error(
                f"There was an error while trying to start the session.\n{response}"
            )
            return False

        self.log.info(f"Session {session_id} has been successfully started")
        return True

    def initiate_session(
        self,
        name: str,
        track_ids: list[str],
        assign_to: list[str] = [],
        session_config: SessionConfig = SessionConfig(),
        description: str = "",
    ) -> tuple[str | None, list[str] | None]:
        """Create, assign cases and start a session.

        Args:
            name (str): Name of a session.
            track_ids (list[str]): List of cases' track IDs.
            assign_to (list[str]): List of users' emails that will be
            assigned as the session evaluators. Session is always assigned
            to the owner, even when his email is not provided. Defaults to
            [].
            session_config (SessionConfig, optional): Session configuration.
            Defaults to SessionConfig() (default configuration).
            description (str, optional): Session description. Defaults to "".

        Returns:
            tuple[str, list[str]] | None: Tuple that contains two values. First
            one is a created session ID and a second one is a list of created
            session case ID from assigned cases. Returns `tuple[None, None]` if
            there was an error during session initiation.
        """
        if not track_ids:
            self.log.error("Session requires at least one case to be assigned to it!")
            return None, None

        response, success, _ = self.http_client.get(CATWALK_GET_USERS_ENDPOINT)

        if not success:
            self.log.error(f"There was an error while fetching users list.\n{response}")
            return None, None

        users = loads(response)["users"]
        users = [user["id"] for user in users if user["email"] in assign_to]

        response, success, _ = self.http_client.post(
            "/api/evaluation/sessions/initiate",
            {
                "name": name,
                "description": description,
                "track_ids": track_ids,
                "assigned_users": users,
                "config": session_config.dict(),
            },
        )

        if not success:
            self.log.error(
                f"There was an error while trying to initiate a session.\n{response}"
            )
            return None, None

        res = loads(response)
        session_id = res["session_id"]
        session_case_ids = res["session_case_ids"]
        self.log.info(f"Session {session_id} has been successfully initiated")
        return session_id, session_case_ids



from os import environ
from json import loads

from urllib.request import Request

from catwalk_client.common.exception import CatwalkClientException

from ._http_client import HTTPClient
from .constants import (
    CATWALK_AUTH_HEADER,
    CATWALK_USER_AGENT_HEADER_VALUE,
    CATWALK_CLIENT_LOCATION,
)


class CatwalkHTTPClient(HTTPClient):
    """Custom Catwalk HTTP client."""

    def __init__(
        self,
        catwalk_url: str,
        auth_token: str,
        insecure: bool = True,
        timeout: int = 30,
        timezone: str = "UTC",
    ):
        """Initialize a `CatwalkHTTPClient` object.

        Args:
            catwalk_url (str): URL of a Catwalk instance.
            auth_token (str): Catwalk authorization token.
            insecure (bool, optional): Whether to use insecure SSL mode.
            Defaults to True.
            timeout (int, optional): Request timeout (in seconds).
            Defaults to 30.
            timezone (str, optional): Timezone name. This value will
            automatically be applied to a `Client-Location` header.
            Defaults to "UTC".
        """
        super().__init__(
            catwalk_url or environ.get("CATWALK_URL"),
            auth_token or environ.get("CATWALK_AUTH_TOKEN"),
            insecure,
            timeout,
        )

        self.timezone = timezone

    def _apply_headers(self, request: Request):
        """Apply default headers.

        Applies `User-Agent`, `Catwalk-Authorization` and `Client-Location`
        headers.

        Args:
            request (Request): `urllib`'s Request object.
        """
        self._add_client_location_header(request)
        self._add_auth_token_header(request, self.auth_token)
        self._add_user_agent_header(request, CATWALK_USER_AGENT_HEADER_VALUE)

    def _add_auth_token_header(self, request: Request, header_value: str = ""):
        """Apply a `Catwalk-Authorization` header to a given request.

        Args:
            request (Request): `urllib`'s Request object.
            header_value (str, optional): Bearer token value. Defaults
            to "".
        """
        request.add_header(CATWALK_AUTH_HEADER, f"Bearer {header_value}")

    def _add_client_location_header(self, request: Request):
        """Apply a `Client-Location` header to a given request.

        Args:
            request (Request): `urllib`'s Request object.
        """
        request.add_header(CATWALK_CLIENT_LOCATION, self.timezone)

    def fetch_auth_token(self, email: str, password: str) -> str:
        """Get authorization token for a given credentials.

        Args:
            email (str): Email of a Catwalk user account.
            password (str): Password to a Catwalk user account.

        Raises:
            CatwalkClientException: On unsuccessful request call.

        Returns:
            str: User's authorization token.
        """
        response, success, _ = self.post(
            "/api/auth/login", {"email": email, "password": password}
        )

        if not success:
            raise CatwalkClientException(response)

        return loads(response)["token"]



from __future__ import annotations

from json import loads, dumps
from ssl import SSLContext, CERT_NONE, create_default_context
from typing import Any, Mapping, Sequence, Tuple, Union

from urllib.error import URLError, HTTPError
from urllib.parse import urlencode
from urllib.request import urlopen, Request

QueryType = Union[
    Mapping[Any, Any],
    Mapping[Any, Sequence[Any]],
    Sequence[Tuple[Any, Any]],
    Sequence[Tuple[Any, Sequence[Any]]],
]


class HTTPClient:
    """HTTP client that uses `urllib` to execute requests."""

    def __init__(
        self, url: str, auth_token: str, insecure: bool = True, timeout: int = 30
    ):
        """Initialize HTTPClient object.

        Args:
            url (str): URL to which future request will be made.
            auth_token (str): Bearer authorization token.
            insecure (bool, optional): Whether to use insecure SSL mode.
            Defaults to True.
            timeout (int, optional): Request timeout (in seconds).
            Defaults to 30.
        """
        self.url = url
        self.auth_token = auth_token
        self.insecure = insecure
        self.timeout = timeout

    def get_url(self, path: str) -> str:
        """Return full URL.

        Args:
            path (str): URL path (must begin with `/`).

        Returns:
            str: Full URL path e.g. `http://localhost/test-endpoint` where
            `/test-endpoint` is a given path string and `http://localhost`
            is client's configuration URL.
        """
        return self.url.rstrip("/") + path

    def get(
        self,
        url_postfix: str,
        query_params: QueryType | None = None,
        doseq: bool = False,
    ) -> tuple[str, bool, int]:
        """Makes an HTTP GET call.

        Args:
            url_postfix (str): URL endpoint path.
            query_params (QueryType | None, optional): URL query params.
            Defaults to None.
            doseq (bool, optional): If any values in the query arg are
            sequences and `doseq` is `True`, each sequence element is
            converted to a separate parameter. Defaults to False.

        Returns:
            tuple[str, bool, int]: Returns three values in form of a tuple:
            response data/message in form of a string, boolean value that
            informs if the request was executed successfully and a response
            status code.
        """
        url = self.get_url(url_postfix) + self._encode_query_params(query_params, doseq)
        req = Request(url=url)
        return self._make_request(req)

    def post(
        self,
        url_postfix: str,
        payload: dict | None,
        query_params: QueryType | None = None,
        doseq: bool = False,
    ) -> tuple[str, bool, int]:
        """Makes an HTTP POST call.

        Args:
            url_postfix (str): URL endpoint path.
            payload (dict | None): Request's body.
            query_params (QueryType | None, optional): URL query params.
            Defaults to None.
            doseq (bool, optional): If any values in the query arg are
            sequences and `doseq` is `True`, each sequence element is
            converted to a separate parameter. Defaults to False.

        Returns:
            tuple[str, bool, int]: Returns three values in form of a tuple:
            response data/message in form of a string, boolean value that
            informs if the request was executed successfully and a response
            status code.
        """
        req_data = dumps(payload).encode() if payload else b""
        url = self.get_url(url_postfix) + self._encode_query_params(query_params, doseq)
        req = Request(url=url, data=req_data)
        req.add_header("Content-Type", "application/json")
        return self._make_request(req)

    def patch(
        self,
        url_postfix: str,
        payload: dict | None,
        query_params: QueryType | None = None,
        doseq: bool = False,
    ) -> tuple[str, bool, int]:
        """Makes an HTTP PATCH call.

        Args:
            url_postfix (str): URL endpoint path.
            payload (dict | None): Request's body.
            query_params (QueryType | None, optional): URL query params.
            Defaults to None.
            doseq (bool, optional): If any values in the query arg are
            sequences and `doseq` is `True`, each sequence element is
            converted to a separate parameter. Defaults to False.

        Returns:
            tuple[str, bool, int]: Returns three values in form of a tuple:
            response data/message in form of a string, boolean value that
            informs if the request was executed successfully and a response
            status code.
        """
        req_data = dumps(payload).encode() if payload else b""
        url = self.get_url(url_postfix) + self._encode_query_params(query_params, doseq)
        req = Request(url=url, data=req_data, method="PATCH")
        req.add_header("Content-Type", "application/json")
        return self._make_request(req)

    def _make_request(self, request: Request) -> tuple[str, bool, int]:
        """Execute given request.

        Args:
            request (Request): `urllib`'s Request object.

        Returns:
            tuple[str, bool, int]: Returns three values in form of a tuple:
            response data/message in form of a string, boolean value that
            informs if the request was executed successfully and a response
            status code.
        """
        self._apply_headers(request)
        try:
            resp = urlopen(
                request, timeout=self.timeout, context=self._get_ssl_context()
            )
        except HTTPError as e:
            e_body = self._parse_http_error(e)
            return (
                f"Error: {e}\n[URL] {e.geturl()}\n[DETAILS] {e_body}",
                False,
                e.getcode(),
            )
        except URLError as e:
            return f"Connection error: {e.reason}", False, -1
        except ConnectionResetError as e:
            return f"Connection reset error: {e}", False, -1
        except Exception as e:
            return f"Error: {e}", False, -1

        return resp.read().decode(), True, resp.getcode()

    def _parse_http_error(self, error: HTTPError) -> str | dict | None:
        """Parse HTTP response error body. Handles `text/html` and `application/json`.

        Args:
            error (HTTPError): Request response error.

        Returns:
            str | dict | None: Decoded response error body. Returns None if
            `Content-Type` of given response isn't either a `text/html` or
            an `application/json`.
        """
        content_type = error.headers.get_content_type()

        if content_type == "application/json":
            return self._parse_http_error_json(error)
        elif content_type == "text/html":
            return self._parse_http_error_html(error)

        return None

    def _parse_http_error_json(self, error: HTTPError) -> dict:
        """Parse HTTP error with `Content-Type: application/json`.

        Args:
            error (HTTPError): Request response error.

        Returns:
            dict: Decoded response error JSON body.
        """
        return loads(error.read().decode())

    def _parse_http_error_html(self, error: HTTPError) -> str:
        """Parse HTTP error with `Content-Type: text/html`.

        Args:
            error (HTTPError): Request response error.

        Returns:
            str: Decoded response error HTML body.
        """
        return error.read().decode()

    def _get_ssl_context(self) -> SSLContext:
        """Get request's SSL context. If `HTTPClient.insecure` is set to
        `True` it won't validate the hostname and the certificate.

        Returns:
            SSLContext: SSL context.
        """
        ctx = create_default_context()
        if self.insecure:
            ctx.check_hostname = False
            ctx.verify_mode = CERT_NONE
        return ctx

    def _encode_query_params(
        self, query_params: QueryType | None, doseq: bool = False
    ) -> str:
        """Encode HTTP query parameters.

        Args:
            query_params (QueryType | None): Query parameters in form of
            a mapping or sequence.
            doseq (bool, optional): If any values in the query arg are
            sequences and `doseq` is `True`, each sequence element is
            converted to a separate parameter. Defaults to False.

        Returns:
            str: Encoded query parameters in form of a string. It
            automatically adds the `?` as a prefix.
        """
        if query_params is None or len(query_params) == 0:
            return ""
        encoded_params = urlencode(query=query_params, doseq=doseq)
        return "?" + encoded_params

    def _apply_headers(self, request: Request):
        """Apply default headers.

        Applies `User-Agent` and `Authorization` headers.

        Args:
            request (Request): `urllib`'s Request object.
        """
        self._add_user_agent_header(request)
        self._add_auth_token_header(request, self.auth_token)

    def _add_auth_token_header(self, request: Request, header_value: str = ""):
        """Apply an `Authorization` header to a given request.

        Args:
            request (Request): `urllib`'s Request object.
            header_value (str, optional): Bearer token value.
            Defaults to "".
        """
        request.add_header("Authorization", f"Bearer {header_value}")

    def _add_user_agent_header(
        self, request: Request, header_value: str = "Request/1.0"
    ):
        """Apply a `User-Agent` header to a given request.

        Args:
            request (Request): `urllib`'s Request object.
            header_value (str, optional): User agent information.
            Defaults to "Request/1.0".
        """
        if not request.has_header("User-Agent"):
            request.add_header("User-Agent", header_value)
